#include "ops.h"

#define dprintln(a...) do { if (debug_mode&DEBUG_MARSHAL) println(a); } while (0)

CBEMarshalOp *CBEObjectType::buildMarshalOps(CMSConnection *connection, int channels, CASTExpression *rvalue, const char *name, CBEType *originalType, CBEParameter *param, int flags)

{
  CBEMarshalOp *result;

  dprintln("Marshalling %s (%s)", name, aoi->name);
  indent(+1);

  result = new CBEOpSimpleCopy(connection, channels, rvalue, originalType, getFlatSize(), name, param, flags);
  
  indent(-1);
  
  return result;
}

CASTDeclaration *CBEObjectType::buildDeclaration(CASTDeclarator *decl, CASTCompoundStatement *compound)

{
  return new CASTDeclaration(new CASTTypeSpecifier("CORBA_Object"), decl, compound);
}

int CBEObjectType::getArgIndirLevel(CBEDeclType declType)

{ 
  if ((declType==INOUT) || (declType==OUT))
    return 1;

  return 0;    
}

CASTStatement *CBEObjectType::buildTestClientInit(CASTExpression *globalPath, CASTExpression *localPath, CBEVarSource *varSource, CBEDeclType type, CBEParameter *param)

{
  if ((type!=IN) && (type!=INOUT))
    return NULL;

  CASTExpression *value = new CASTFunctionOp(
    new CASTIdentifier("idl4_get_random_threadid")
  );
  
  if (param->aoi->getProperty("handle"))
    value = new CASTIdentifier("server");

  return new CASTExpressionStatement(
    new CASTBinaryOp("=",
      knitPathInstance(globalPath, "svr_in"),
      new CASTBinaryOp("=",
        knitPathInstance(globalPath, "client"),
        value))
  );
}

CASTStatement *CBEObjectType::buildTestClientCheck(CASTExpression *globalPath, CASTExpression *localPath, CBEVarSource *varSource, CBEDeclType type)

{
  if ((type!=INOUT) && (type!=OUT))
    return NULL;

  char info[100];

  enableStringMode(info, sizeof(info));
  localPath->write();
  disableStringMode();

  return new CASTIfStatement(
    new CASTUnaryOp("!", 
      new CASTFunctionOp(
        new CASTIdentifier("idl4_threads_equal"),
        knitExprList(
          knitPathInstance(globalPath, "client"),
          knitPathInstance(globalPath, "svr_out")))),
    new CASTExpressionStatement(
      new CASTFunctionOp(
        new CASTIdentifier("panic"),
        knitExprList(new CASTStringConstant(false, mprintf("Out transfer fails for %s: %%d!=%%d", info)),
                     knitPathInstance(globalPath, "client")->clone(),
                     knitPathInstance(globalPath, "svr_out")->clone())))
  );
}

CASTStatement *CBEObjectType::buildTestServerInit(CASTExpression *globalPath, CASTExpression *localPath, CBEVarSource *varSource, CBEDeclType type, CBEParameter *param, bool bufferAvailable)

{
  if ((type!=INOUT) && (type!=OUT))
    return NULL;

  return new CASTExpressionStatement(
    new CASTBinaryOp("=",
      knitPathInstance(globalPath, "svr_out"),
      new CASTBinaryOp("=",
        localPath,
        new CASTFunctionOp(
          new CASTIdentifier("idl4_get_random_threadid"))))
  );
}

CASTStatement *CBEObjectType::buildTestServerCheck(CASTExpression *globalPath, CASTExpression *localPath, CBEVarSource *varSource, CBEDeclType type)

{
  if ((type!=IN) && (type!=INOUT))
    return NULL;

  char info[100];

  enableStringMode(info, sizeof(info));
  localPath->write();
  disableStringMode();

  return new CASTIfStatement(
    new CASTUnaryOp("!",
      new CASTFunctionOp(
        new CASTIdentifier("idl4_threads_equal"),
        knitExprList(
          localPath->clone(),
          knitPathInstance(globalPath, "svr_in")))),
    new CASTExpressionStatement(
      new CASTFunctionOp(
        new CASTIdentifier("panic"),
        knitExprList(new CASTStringConstant(false, mprintf("In transfer fails for %s: %%d!=%%d", info)),
                     localPath->clone(),
                     knitPathInstance(globalPath, "svr_in"))))
  );
}

CASTStatement *CBEObjectType::buildTestServerRecheck(CASTExpression *globalPath, CASTExpression *localPath, CBEVarSource *varSource, CBEDeclType type)

{
  if (type!=IN)
    return NULL;

  char info[100];

  enableStringMode(info, sizeof(info));
  localPath->write();
  disableStringMode();

  return new CASTIfStatement(
    new CASTUnaryOp("!",
      new CASTFunctionOp(
        new CASTIdentifier("idl4_threads_equal"),
        knitExprList(
          localPath->clone(),
          knitPathInstance(globalPath, "svr_in")))),
    new CASTExpressionStatement(
      new CASTFunctionOp(
        new CASTIdentifier("panic"),
        knitExprList(new CASTStringConstant(false, mprintf("Output value collision for %s: %%d!=%%d", info)),
                     localPath->clone(),
                     knitPathInstance(globalPath, "svr_in"))))
  );
}

CASTExpression *CBEObjectType::buildDefaultValue()

{
  return new CASTIdentifier("idl4_nilthread");
}

CASTStatement *CBEObjectType::buildTestDisplayStmt(CASTExpression *globalPath, CASTExpression *localPath, CBEVarSource *varSource, CBEDeclType type, CASTExpression *value)

{
  return new CASTExpressionStatement(
    new CASTFunctionOp(
      new CASTIdentifier("printf"),
      knitExprList(
        new CASTStringConstant(false, "%t"),
        value))
  );
}

int CBEObjectType::getAlignment()

{
  return msFactory->getThreadIDType()->getAlignment();
}

int CBEObjectType::getFlatSize()

{
  return msFactory->getThreadIDType()->getFlatSize();
}
